package models

import (
	"encoding/json"
	"net/http"
	"strconv"
	"strings"

	"forge.grandlyon.com/systemes-dinformation/project-template/sdk-go/internal/auth"
)

// HandleBankAccounts handle API call on BankAccounts
func (d *DataHandler) HandleBankAccounts(w http.ResponseWriter, r *http.Request) {
	id, _ := strconv.Atoi(strings.TrimPrefix(r.URL.Path, "/api/BankAccounts/"))
	switch method := r.Method; method {
	case "GET":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.getBankAccountAdmin(w, r, id)
		case "BANKER":
			d.getBankAccountBanker(w, r, id)
		case "CLIENT":
			d.getBankAccountClient(w, r, id)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "POST":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.postBankAccountAdmin(w, r, id)
		case "BANKER":
			d.postBankAccountBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "PUT":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.putBankAccountAdmin(w, r, id)
		case "BANKER":
			d.putBankAccountBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "DELETE":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.deleteBankAccountAdmin(w, r, id)
		case "BANKER":
			d.deleteBankAccountBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	default:
		http.Error(w, "method not allowed", 400)
	}
}

func (d *DataHandler) getBankAccountAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o BankAccount
		if err := d.db.Preload("Operations").First(&o, id).Error; err != nil {
			http.Error(w, ErrorIDDoesNotExist, http.StatusNotFound)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []BankAccount
		d.db.Preload("Operations").Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) getBankAccountBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	if id != 0 {
		var o BankAccount
		if err := d.db.Preload("Operations").First(&o, id).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		// Check that the bank account belong to a one of the banker's client
		var userClient UserClient
		if err := d.db.Where(reqIDAndBankerID, o.UserClientID, user.ID).First(&userClient).Error; err != nil {
			http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []BankAccount
		d.db.Preload("Operations").Where("user_client_id IN (?)", d.db.Table("user_clients").Select("id").Where("user_banker_id = ?", user.ID).QueryExpr()).Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) getBankAccountClient(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserClient)
	if id != 0 {
		var o BankAccount
		if err := d.db.Preload("Operations").Where("id = ? AND user_client_id = ?", id, user.ID).First(&o).Error; err != nil {
			http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []BankAccount
		d.db.Preload("Operations").Where("user_client_id = ?", user.ID).Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) postBankAccountAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o BankAccount
	err := json.NewDecoder(r.Body).Decode(&o)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	if o.UserClientID == 0 {
		http.Error(w, ErrorUserIDIsMissing, http.StatusInternalServerError)
	}
	d.db.Create(&o)
	d.db.Last(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) postBankAccountBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	var o BankAccount
	err := json.NewDecoder(r.Body).Decode(&o)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}

	if o.UserClientID == 0 {
		http.Error(w, ErrorUserIDIsMissing, http.StatusInternalServerError)

	}
	var userClient UserClient
	if err := d.db.Where(reqIDAndBankerID, o.UserClientID, user.ID).First(&userClient).Error; err != nil {
		http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
		return
	}

	d.db.Create(&o)
	d.db.Last(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) putBankAccountAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o BankAccount
	if err := d.db.Preload("Operations").First(&o, id).Error; err != nil {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
		return
	}
	var bankAccount BankAccount
	err := json.NewDecoder(r.Body).Decode(&bankAccount)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.BankOverdraft = bankAccount.BankOverdraft
	o.Type = bankAccount.Type
	if o.UserClientID == 0 {
		http.Error(w, ErrorUserIDIsMissing, http.StatusInternalServerError)
	}
	d.db.Save(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) putBankAccountBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	var o BankAccount
	if err := d.db.Preload("Operations").First(&o, id).Error; err != nil {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
		return
	}
	var bankAccount BankAccount
	err := json.NewDecoder(r.Body).Decode(&bankAccount)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.BankOverdraft = bankAccount.BankOverdraft
	o.Type = bankAccount.Type
	if o.UserClientID == 0 {
		http.Error(w, ErrorUserIDIsMissing, http.StatusInternalServerError)
	}
	var userClient UserClient
	if err := d.db.Where(reqIDAndBankerID, o.UserClientID, user.ID).First(&userClient).Error; err != nil {
		http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
		return
	}
	d.db.Save(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) deleteBankAccountAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o BankAccount
		if err := d.db.First(&o, id).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		d.db.Delete(&o)
	} else {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
	}
}

func (d *DataHandler) deleteBankAccountBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	if id != 0 {
		var o BankAccount
		if err := d.db.First(&o, id).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		var userClient UserClient
		if err := d.db.Where(reqIDAndBankerID, o.UserClientID, user.ID).First(&userClient).Error; err != nil {
			http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
			return
		}
		d.db.Delete(&o)
	} else {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
	}
}
