package models

import (
	"encoding/json"
	"net/http"
	"strconv"
	"strings"

	"forge.grandlyon.com/systemes-dinformation/project-template/sdk-go/internal/auth"
)

// HandleBankers handle API calls on Bankers
func (d *DataHandler) HandleBankers(w http.ResponseWriter, r *http.Request) {
	id, _ := strconv.Atoi(strings.TrimPrefix(r.URL.Path, "/api/UserBankers/"))
	switch method := r.Method; method {
	case "GET":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.getBankerAdmin(w, r, id)
		case "BANKER":
			d.getBankerBanker(w, r, id)
		case "CLIENT":
			d.getBankerClient(w, r, id)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "POST":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.postBankerAdmin(w, r, id)
		case "BANKER", "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}

	case "PUT":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.putBankerAdmin(w, r, id)
		case "BANKER", "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "DELETE":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.deleteBankerAdmin(w, r, id)
		case "BANKER", "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}

	default:
		http.Error(w, "method not allowed", 400)
	}
}

func (d *DataHandler) getBankerAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o UserBanker
		if err := d.db.Preload("UserClients").Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []UserBanker
		d.db.Preload("UserClients").Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) getBankerBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	if id != 0 {
		var o UserBanker
		if err := d.db.Preload("UserClients").Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		if o.ID != user.ID {
			http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
	}
}

func (d *DataHandler) getBankerClient(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserClient)
	if id != 0 && int(user.ID) == id {
		var userClient UserClient
		if err := d.db.Where(reqUserID, user.ID).First(&userClient).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		var o UserBanker
		if err := d.db.Where(reqID, userClient.UserBankerID).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
	}
}

func (d *DataHandler) postBankerAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o UserBanker
	err := json.NewDecoder(r.Body).Decode(&o)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	// check userID is not already present in DB
	var client UserClient
	if err := d.db.Where(reqUserID, o.UserID).First(&client).Error; err == nil {
		http.Error(w, "UserID is already bind to a Client", http.StatusInternalServerError)
		return
	}
	var banker UserBanker
	if err := d.db.Where(reqUserID, o.UserID).First(&banker).Error; err == nil {
		http.Error(w, "UserID is already bind to a Banker", http.StatusInternalServerError)
		return
	}
	d.db.Create(&o)
	d.db.Last(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) putBankerAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o UserBanker
	if err := d.db.Preload("UserClients").Where(reqID, id).First(&o).Error; err != nil {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
		return
	}
	var updatedBanker UserBanker
	err := json.NewDecoder(r.Body).Decode(&updatedBanker)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.Name = updatedBanker.Name
	d.db.Save(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) deleteBankerAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o UserBanker
		if err := d.db.First(&o, id).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		d.db.Delete(&o)
	} else {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
	}
}
