package models

import (
	"encoding/json"
	"net/http"
	"strconv"
	"strings"

	"forge.grandlyon.com/systemes-dinformation/project-template/sdk-go/internal/auth"
)

// HandleClients expose the UserClients API
func (d *DataHandler) HandleClients(w http.ResponseWriter, r *http.Request) {
	id, _ := strconv.Atoi(strings.TrimPrefix(r.URL.Path, "/api/UserClients/"))
	switch method := r.Method; method {
	case "GET":

		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.getClientAdmin(w, r, id)
		case "BANKER":
			d.getClientBanker(w, r, id)
		case "CLIENT":
			d.getClientClient(w, r, id)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "POST":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.postClientAdmin(w, r, id)
		case "BANKER":
			d.postClientBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "PUT":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.putClientAdmin(w, r, id)
		case "BANKER":
			d.putClientBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	case "DELETE":
		switch auth.GetLoggedUserTechnical(w, r).Role {
		case "ADMIN":
			d.deleteClientAdmin(w, r, id)
		case "BANKER":
			d.deleteClientBanker(w, r, id)
		case "CLIENT":
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
		default:
			http.Error(w, ErrorRoleOfLoggedUser, http.StatusInternalServerError)
		}
	default:
		http.Error(w, "method not allowed", 400)
	}
}

func (d *DataHandler) getClientAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o UserClient
		if err := d.db.Preload("BankAccounts").Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []UserClient
		d.db.Preload("BankAccounts").Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) getClientBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	if id != 0 {
		var o UserClient
		if err := d.db.Preload("BankAccounts").Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		if o.UserBankerID != user.ID {
			http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else {
		var o []UserClient
		d.db.Preload("BankAccounts").Where("user_banker_id = ?", user.ID).Find(&o)
		json.NewEncoder(w).Encode(o)
	}
}

func (d *DataHandler) getClientClient(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserClient)
	if id != 0 && int(user.ID) == id {
		var o UserClient
		if err := d.db.Preload("BankAccounts").Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		json.NewEncoder(w).Encode(o)
	} else if id == 0 {
		var o []UserClient
		d.db.Preload("BankAccounts").Where(reqID, user.ID).Find(&o)
		json.NewEncoder(w).Encode(o)
	} else {
		http.Error(w, ErrorCannotAccessRessource, http.StatusForbidden)
		return
	}
}

func (d *DataHandler) postClientAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o UserClient
	err := json.NewDecoder(r.Body).Decode(&o)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	// check userID is not already present in DB
	var client UserClient
	if err := d.db.Where(reqUserID, o.UserID).First(&client).Error; err == nil {
		http.Error(w, "UserID is already bind to a Client", http.StatusInternalServerError)
		return
	}
	var banker UserBanker
	if err := d.db.Where(reqUserID, o.UserID).First(&banker).Error; err == nil {
		http.Error(w, "UserID is already bind to a Banker", http.StatusInternalServerError)
		return
	}
	d.db.Create(&o)
	d.db.Last(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) postClientBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	var o UserClient
	err := json.NewDecoder(r.Body).Decode(&o)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.UserBankerID = user.ID
	d.db.Create(&o)
	d.db.Last(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) putClientAdmin(w http.ResponseWriter, r *http.Request, id int) {
	var o UserClient
	if err := d.db.Preload("BankAccounts").Where(reqID, id).First(&o).Error; err != nil {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
		return
	}
	var updatedUser UserClient
	err := json.NewDecoder(r.Body).Decode(&updatedUser)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.Name = updatedUser.Name
	o.UserBankerID = updatedUser.UserBankerID
	d.db.Save(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) putClientBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	var o UserClient
	if err := d.db.Preload("BankAccounts").Where(reqID, id).First(&o).Error; err != nil {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
		return
	}
	var updatedUser UserClient
	err := json.NewDecoder(r.Body).Decode(&updatedUser)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	o.Name = updatedUser.Name
	o.UserBankerID = user.ID
	d.db.Save(&o)
	json.NewEncoder(w).Encode(o)
}

func (d *DataHandler) deleteClientAdmin(w http.ResponseWriter, r *http.Request, id int) {
	if id != 0 {
		var o UserClient
		if err := d.db.Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		// Delete bank accounts of user
		d.db.Where("user_client_id = ?", o.ID).Delete(&BankAccount{})

		d.db.Delete(&o)
	} else {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
	}
}

func (d *DataHandler) deleteClientBanker(w http.ResponseWriter, r *http.Request, id int) {
	user := d.getLoggedUser(w, r).(UserBanker)
	if id != 0 {
		var o UserClient
		if err := d.db.Where(reqID, id).First(&o).Error; err != nil {
			http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
			return
		}
		if o.UserBankerID != user.ID {
			http.Error(w, ErrorNotAuthorizeMethodOnRessource, http.StatusMethodNotAllowed)
			return
		}
		// Delete bank accounts of user
		d.db.Where("user_client_id = ?", o.ID).Delete(&BankAccount{})

		d.db.Delete(&o)
	} else {
		http.Error(w, ErrorIDIsMissing, http.StatusNotFound)
	}
}
