package models

import (
	"net/http"
	"strings"
	"time"

	"forge.grandlyon.com/systemes-dinformation/project-template/sdk-go/internal/auth"
	"github.com/jinzhu/gorm"

	// Needed for sqlite
	_ "github.com/jinzhu/gorm/dialects/sqlite"
)

// DataHandler init a gorm DB an presents API handlers
type DataHandler struct {
	db *gorm.DB
}

// ErrorIDDoesNotExist = "id does not exist" with 404 http.StatusNotFound
const ErrorIDDoesNotExist = "id does not exist"

// ErrorIDIsMissing = "id is missing" with 404 http.StatusNotFound
const ErrorIDIsMissing = "id is missing"

// ErrorCannotAccessRessource = "You can not access this ressource" with 403 http.StatusForbidden
const ErrorCannotAccessRessource = "You can not access this ressource"

// ErrorRoleOfLoggedUser = "Could not get role of logged user" with 500 http.StatusInternalServerError
const ErrorRoleOfLoggedUser = "Could not get role of logged user"

// ErrorNotAuthorizeMethodOnRessource = "You're not authorize to execute this method on this ressource." with 405 http.StatusMethodNotAllowed
const ErrorNotAuthorizeMethodOnRessource = "You're not authorize to execute this method on this ressource."

// ErrorUserIDIsMissing = "id of UserClient is missing"
const ErrorUserIDIsMissing = "id of UserClient is missing"

const reqID = "id = ?"
const reqUserID = "user_id = ?"
const reqIDAndBankerID = "id = ? and user_banker_id = ?"

// NewDataHandler init a DataHandler and returns a pointer to it
func NewDataHandler() *DataHandler {
	db, err := gorm.Open("sqlite3", "./data/test.db")
	if err != nil {
		panic("failed to connect database")
	}
	db.LogMode(true)

	// Migrate the schema
	db.AutoMigrate(&UserClient{})
	db.AutoMigrate(&UserBanker{})
	db.AutoMigrate(&BankAccount{})
	db.AutoMigrate(&Operation{})
	return &DataHandler{db: db}
}

// ProcessAPI redirect API call to DataHandler to each correct API func
func (d *DataHandler) ProcessAPI(w http.ResponseWriter, r *http.Request) {
	api := strings.Split(strings.TrimPrefix(r.URL.Path, "/api/"), "/")[0]
	switch api {
	case "UserClients":
		d.HandleClients(w, r)
	case "UserBankers":
		d.HandleBankers(w, r)
	case "BankAccounts":
		d.HandleBankAccounts(w, r)
	case "Operations":
		d.HandleOperations(w, r)
	}
}

// UserBanker has many UserClient
type UserBanker struct {
	ID          uint       `gorm:"primary_key"`
	CreatedAt   time.Time  `json:"-"`
	UpdatedAt   time.Time  `json:"-"`
	DeletedAt   *time.Time `json:"-"`
	UserID      int        `gorm:"not null;unique"`
	Name        string
	UserClients []UserClient
}

// UserClient has many BankAccounts and one UserBanker
type UserClient struct {
	ID           uint       `gorm:"primary_key"`
	CreatedAt    time.Time  `json:"-"`
	UpdatedAt    time.Time  `json:"-"`
	DeletedAt    *time.Time `json:"-"`
	UserID       int        `gorm:"not null;unique"`
	Name         string
	UserBankerID uint
	BankAccounts []BankAccount
}

// BankAccount belongs to an UserClient
type BankAccount struct {
	ID            uint       `gorm:"primary_key"`
	CreatedAt     time.Time  `json:"-"`
	UpdatedAt     time.Time  `json:"-"`
	DeletedAt     *time.Time `json:"-"`
	Number        string
	UserClientID  uint
	Type          string
	Amount        int
	BankOverdraft int
	Operations    []Operation `gorm:"foreignkey:Debtor"`
}

// Operation reprensent a bank transaction registered in a BankAccount
type Operation struct {
	ID        uint `gorm:"primary_key"`
	Debtor    uint
	CreatedAt time.Time  `json:"-"`
	UpdatedAt time.Time  `json:"-"`
	DeletedAt *time.Time `json:"-"`
	Amount    int
	Date      time.Time
	Creditor  uint
}

func (d *DataHandler) getLoggedUser(w http.ResponseWriter, r *http.Request) interface{} {
	user := auth.GetLoggedUserTechnical(w, r)
	if user.Role != "" && (user.Role == "BANKER") {
		var o UserBanker
		if err := d.db.Where(reqUserID, user.ID).First(&o).Error; err != nil {
			o := UserBanker{UserID: user.ID, Name: user.Login}
			d.db.Create(&o)
			d.db.Where(reqUserID, user.ID).First(&o)
			return o
		}
		return o
	} else if user.Role != "" && (user.Role == "CLIENT") {
		var o UserClient
		if err := d.db.Where(reqUserID, user.ID).First(&o).Error; err != nil {
			o := UserClient{UserID: user.ID, Name: user.Login}
			d.db.Create(&o)
			d.db.Where(reqUserID, user.ID).First(&o)
			return o
		}

		return o
	}

	return nil
}
