// Imports
import * as Auth from "/services/auth/auth.js";
import * as Messages from "/services/messages/messages.js";
import * as Users from "/components/users/users.js";

// DOM elements

// local variables
let current_user;

export async function mount(where, parent) {
  const operation = new Operation(parent);
  await operation.mount(where);
  operation.handleDom();
  return operation;
}

class Operation {
  constructor(parent) {
    this.parent = parent;
  }

  async mount(where) {
    current_user = await Auth.GetUser();

    document.getElementById(where).innerHTML = /* HTML */ `
      <div class="table-container">
        <table class="table is-bordered is-narrow is-hoverable is-fullwidth">
          <thead>
            <tr class="is-selected">
              <th>Id</th>
              <th>Montant</th>
              <th>Compte bancaire</th>
              <th>Date</th>
              ${current_user.role != "CLIENT" ? `<th>Actions</th>` : ``}
            </tr>
          </thead>
          <tbody id="operations-table"></tbody>
        </table>
      </div>
      <nav class="level">
        <div class="level-left"></div>
        <!-- Right side -->
        <div class="level-right">
          <button id="operation-new" class="button is-success">
            <span class="icon is-small">
              <i class="fas fa-plus"></i>
            </span>
          </button>
        </div>
      </nav>

      <div class="modal" id="operation-modal">
        <div class="modal-background"></div>
        <div class="modal-card" id="operation-modal-card">
          <header class="modal-card-head">
            <p class="modal-card-title">Ajout d'une opération</p>
            <button class="delete" aria-label="close" id="operation-modal-close"></button>
          </header>
          <section class="modal-card-body">
            <div class="field">
              <label>Montant</label>
              <div class="control">
                <input class="input" type="text" id="operation-modal-amount" />
              </div>
            </div>
            <div class="field">
              <label>Compte bancaire</label><br />
              <div class="control select">
                <select name="bankAccount" id="operation-modal-bankAccount"> </select>
              </div>
            </div>
          </section>
          <footer class="modal-card-foot">
            <button id="operation-modal-save" class="button is-success">Sauvegarder</button>
            <button id="operation-modal-cancel" class="button">Annuler</button>
          </footer>
        </div>
      </div>
    `;
  }

  handleDom() {
    let operationHandler = this;
    document.getElementById(`operation-new`).addEventListener("click", function () {
      Users.toggleModal("operation-modal", "operation-modal-card");
      operationHandler.newOperation();
    });

    document.getElementById(`operation-modal-close`).addEventListener("click", function () {
      Users.toggleModal("operation-modal", "operation-modal-card");
    });
    document.getElementById(`operation-modal-cancel`).addEventListener("click", function () {
      Users.toggleModal("operation-modal", "operation-modal-card");
    });
    document.getElementById(`operation-modal-save`).addEventListener("click", function () {
      operationHandler.postOperation();
    });
  }

  async displayOperations(bankAccount) {
    if (bankAccount.ID != 0) this.bankAccount = await this.updateBankAccount(bankAccount.ID);
    else this.bankAccount = bankAccount;
    const markup = this.bankAccount.Operations.map((operation) => this.operationTemplate(operation)).join("");
    document.getElementById("operations-table").innerHTML = markup;

    let operationHandler = this;
    if (current_user.role != "CLIENT") {
      this.bankAccount.Operations.map((operation) => {
        document.getElementById(`operations-operation-delete-${operation.ID}`).addEventListener("click", function () {
          operationHandler.deleteOperation(operation);
        });
      });
    }
  }

  operationTemplate(operation) {
    return /* HTML */ `
      <tr id="operations-operation-${operation.ID}">
        <td>${operation.ID}</td>
        <td>${operation.Amount}</td>
        <td>${operation.Creditor}</td>
        <td>${new Date(operation.Date).toLocaleDateString()}</td>
        ${current_user.role != "CLIENT"
          ? `
        <td>
          <a id="operations-operation-delete-${operation.ID}" class="button is-danger is-small">
            <span>Supprimer</span>
            <span class="icon is-small">
              <i class="fas fa-trash"></i>
            </span>
          </a>
        </td>
        `
          : ``}
      </tr>
    `;
  }

  async updateBankAccount(bankAccountID) {
    try {
      const response = await fetch("/api/BankAccounts/" + bankAccountID, {
        method: "GET",
        headers: new Headers({
          "XSRF-Token": current_user.xsrftoken,
        }),
      });
      if (response.status !== 200) {
        throw new Error(`BankAccount could not be fetched (status ${response.status})`);
      }
      return await response.json();
    } catch (e) {
      Messages.Show("is-warning", e.message);
      console.error(e);
    }
  }

  async deleteOperation(operation) {
    try {
      const response = await fetch("/api/Operations/" + operation.ID, {
        method: "delete",
        headers: new Headers({
          "XSRF-Token": current_user.xsrftoken,
        }),
      });
      if (response.status !== 200) {
        throw new Error(`Operation could not be deleted (status ${response.status})`);
      }
      document.getElementById(`operations-operation-${operation.ID}`).remove();
      await this.parent.parent.updateBankAccounts();
      document.getElementById(`accounts-account-edit-${this.bankAccount.ID}`).click();
    } catch (e) {
      Messages.Show("is-warning", e.message);
      console.error(e);
    }
  }

  async newOperation() {
    try {
      const response = await fetch("/api/BankAccounts/", {
        method: "GET",
        headers: new Headers({
          "XSRF-Token": current_user.xsrftoken,
        }),
      });
      if (response.status !== 200) {
        throw new Error(`BankAccounts could not be fetched (status ${response.status})`);
      }
      let bankAccounts = await response.json();

      let select = document.getElementById("operation-modal-bankAccount");
      bankAccounts.forEach(async (bankAccount) => {
        if (bankAccount.ID != this.bankAccount.ID) {
          let el = document.createElement("option");
          let client = await this.getClientById(bankAccount.UserClientID);
          el.textContent = client.Name + " - " + bankAccount.Number;
          el.value = bankAccount.ID;
          select.appendChild(el);
        }
      });
    } catch (e) {
      Messages.Show("is-warning", e.message);
      console.error(e);
    }
  }

  async postOperation() {
    try {
      const response = await fetch("/api/Operations/", {
        method: "POST",
        headers: new Headers({
          "XSRF-Token": current_user.xsrftoken,
        }),
        body: JSON.stringify({
          Debtor: this.bankAccount.ID,
          Creditor: parseInt(document.getElementById("operation-modal-bankAccount").value),
          Amount: parseInt(document.getElementById("operation-modal-amount").value),
        }),
      });
      if (response.status !== 200) {
        throw new Error(`Operation could not be created (status ${response.status})`);
      }
      await this.displayOperations(this.bankAccount);
      await this.parent.parent.updateBankAccounts();
      document.getElementById(`accounts-account-view-${this.bankAccount.ID}`).click();
    } catch (e) {
      Messages.Show("is-warning", e.message);
      console.error(e);
    }
    Users.toggleModal("operation-modal", "operation-modal-card");
  }

  async getClientById(clientID) {
    const response = await fetch("/api/UserClients/" + clientID, {
      method: "GET",
      headers: new Headers({
        "XSRF-Token": current_user.xsrftoken,
      }),
    });
    if (response.status !== 200) {
      throw new Error(`Could not get clients (status ${response.status})`);
    }
    return await response.json();
  }
}
